import time
from collections import defaultdict
from typing import Dict, Mapping

from meddlr.utils import env

_CURRENT_TIMER_STACK = []


def get_timer() -> "AdvancedTimer":
    """
    Returns:
        The :class:`EventStorage` object that's currently being used.
        Throws an error if no :class`EventStorage` is currently enabled.
    """
    assert len(_CURRENT_TIMER_STACK), (
        "get_event_storage() has to be called inside a " "'with EventStorage(...)' context!"
    )
    return _CURRENT_TIMER_STACK[-1]


def time_profile(prefix=None, delimiter=".", disable=None, top_lvl_only=True):
    if disable is None:
        disable = not env.is_profiling_enabled()

    def _decorator(func):
        name = func.__qualname__
        if prefix is not None:
            name = f"{prefix}{delimiter}{name}"

        def _wrapper(*args, **kwargs):
            if disable:
                return func(*args, **kwargs)

            with AdvancedTimer(name, unroll=True) as timer:
                out = func(*args, **kwargs)
                time_args = timer.summarize()

            if (top_lvl_only and len(_CURRENT_TIMER_STACK) == 0) and isinstance(out, Dict):
                if "_profiler" not in out:
                    out["_profiler"] = {}
                out["_profiler"].update(time_args)
            return out

        return _wrapper

    return _decorator


class AdvancedTimer:
    def __init__(self, name, delimiter=".", unroll=False, data=None):
        self.name = name
        self.delimiter = delimiter
        self.unroll = unroll
        self.timer_started = defaultdict(bool)
        if data is None:
            data = {}
        self.data = data

    def start(self, key):
        self.data[key] = time.perf_counter()
        self.timer_started[key] = True

    def time(self, key):
        if not self.timer_started[key]:
            raise ValueError(f"Must start timer for key '{key}' before calling `time`.")
        self.data[key] = time.perf_counter() - self.data[key]
        self.timer_started[key] = False
        return self.data[key]

    def stop(self, key):
        return self.time(key)

    def update(self, data: Mapping):
        self.data.update(data)
        self.timer_started.update({k: False for k in data})

    def summarize(self) -> Dict:
        return {f"{self.name}{self.delimiter}{k}": v for k, v in self.data.items()}

    def __enter__(self):
        _CURRENT_TIMER_STACK.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert _CURRENT_TIMER_STACK[-1] == self
        if self.unroll and len(_CURRENT_TIMER_STACK) > 1:
            _CURRENT_TIMER_STACK[-2].update(self.summarize())
        _CURRENT_TIMER_STACK.pop()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name})"



from meddlr.data.samplers import build, group_sampler, sampler
from meddlr.data.samplers.build import build_train_sampler, build_val_sampler  # noqa: F401
from meddlr.data.samplers.group_sampler import (  # noqa: F401
    AlternatingGroupSampler,
    DistributedGroupSampler,
    GroupSampler,
)
from meddlr.data.samplers.sampler import AlternatingSampler  # noqa: F401

__all__ = []
__all__.extend(build.__all__)
__all__.extend(group_sampler.__all__)
__all__.extend(sampler.__all__)



from .build import build_lr_scheduler, build_optimizer  # noqa: F401
from .lr_scheduler import WarmupCosineLR, WarmupMultiStepLR  # noqa: F401
from .optimizer import GradAccumOptimizer  # noqa: F401

__all__ = [k for k in globals().keys() if not k.startswith("_")]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
from typing import Any, Dict, List

import torch

from meddlr.config import CfgNode

from .lr_scheduler import NoOpLR, WarmupCosineLR, WarmupMultiStepLR
from .optimizer import GradAccumOptimizer


def build_optimizer(cfg: CfgNode, model: torch.nn.Module) -> torch.optim.Optimizer:
    """
    Build an optimizer from config.
    """
    params: List[Dict[str, Any]] = []
    for key, value in model.named_parameters():
        if not value.requires_grad:
            continue
        lr = cfg.SOLVER.BASE_LR
        weight_decay = cfg.SOLVER.WEIGHT_DECAY
        if key.endswith("norm.weight") or key.endswith("norm.bias"):
            weight_decay = cfg.SOLVER.WEIGHT_DECAY_NORM
        elif key.endswith(".bias"):
            lr = cfg.SOLVER.BASE_LR * cfg.SOLVER.BIAS_LR_FACTOR
            weight_decay = cfg.SOLVER.WEIGHT_DECAY_BIAS
        params += [{"params": [value], "lr": lr, "weight_decay": weight_decay}]

    optimizer = _build_opt(params, cfg)
    return optimizer


def _build_opt(params, cfg):
    optim = cfg.SOLVER.OPTIMIZER
    if optim == "SGD":
        optimizer = torch.optim.SGD(params, cfg.SOLVER.BASE_LR, momentum=cfg.SOLVER.MOMENTUM)
    elif optim == "Adam":
        optimizer = torch.optim.Adam(params, cfg.SOLVER.BASE_LR)
    else:
        raise ValueError(f"Optimizer {optim} not supported")

    # Gradient accumulation wrapper.
    num_grad_accum = cfg.SOLVER.GRAD_ACCUM_ITERS
    if num_grad_accum < 1:
        raise ValueError(f"cfg.SOLVER.GRAD_ACCUM_ITERS must be >= 1. Got {num_grad_accum}")
    elif num_grad_accum > 1:
        optimizer = GradAccumOptimizer(optimizer, num_grad_accum)
    return optimizer


def build_lr_scheduler(
    cfg: CfgNode, optimizer: torch.optim.Optimizer
) -> torch.optim.lr_scheduler._LRScheduler:
    """
    Build a LR scheduler from config.
    """
    if not isinstance(optimizer, torch.optim.Optimizer):
        if hasattr(optimizer, "optimizer"):
            optimizer = optimizer.optimizer

    name = cfg.SOLVER.LR_SCHEDULER_NAME
    if not name:
        return NoOpLR(optimizer)
    elif name == "WarmupMultiStepLR":
        return WarmupMultiStepLR(
            optimizer,
            cfg.SOLVER.STEPS,
            cfg.SOLVER.GAMMA,
            warmup_factor=cfg.SOLVER.WARMUP_FACTOR,
            warmup_iters=cfg.SOLVER.WARMUP_ITERS,
            warmup_method=cfg.SOLVER.WARMUP_METHOD,
        )
    elif name == "WarmupCosineLR":
        return WarmupCosineLR(
            optimizer,
            cfg.SOLVER.MAX_ITER,
            warmup_factor=cfg.SOLVER.WARMUP_FACTOR,
            warmup_iters=cfg.SOLVER.WARMUP_ITERS,
            warmup_method=cfg.SOLVER.WARMUP_METHOD,
        )
    elif name in ("StepLR", "WarmupStepLR"):
        if len(cfg.SOLVER.STEPS) != 1:
            raise ValueError("SOLVER.STEPS must have single value for StepLR")
        lr_step = cfg.SOLVER.STEPS[0]
        max_iter = cfg.SOLVER.MAX_ITER
        num_steps = max_iter // lr_step
        steps = [lr_step * x for x in range(1, 1 + num_steps)]

        # No warmup if plain StepLR.
        warmup_iters = cfg.SOLVER.WARMUP_ITERS if name == "WarmupStepLR" else 0
        return WarmupMultiStepLR(
            optimizer,
            steps,
            cfg.SOLVER.GAMMA,
            warmup_factor=cfg.SOLVER.WARMUP_FACTOR,
            warmup_iters=warmup_iters,
            warmup_method=cfg.SOLVER.WARMUP_METHOD,
        )
    else:
        raise ValueError("Unknown LR scheduler: {}".format(name))



import os
import sys
from os import path
from shutil import rmtree

from setuptools import Command, find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


class UploadCommand(Command):
    """Support setup.py upload.
    Adapted from https://github.com/robustness-gym/meerkat.
    """

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(get_version()))
        os.system("git push --tags")

        sys.exit()


def get_version():
    init_py_path = path.join(path.abspath(path.dirname(__file__)), "meddlr", "__init__.py")
    init_py = open(init_py_path, "r").readlines()
    version_line = [l.strip() for l in init_py if l.startswith("__version__")][0]  # noqa: E741
    version = version_line.split("=")[-1].strip().strip("'\"")
    return version


# ---------------------------------------------------
# Setup Information
# ---------------------------------------------------
NAME = "meddlr"
DESCRIPTION = (
    "Meddlr is a config-driven framework built to simplify ML-based "
    "medical image reconstruction and analysis."
)
VERSION = get_version()
AUTHOR = "The Meddlr Team"
EMAIL = "arjundd@stanford.edu"
URL = "https://github.com/ad12/meddlr"
REQUIRES_PYTHON = ">=3.6"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("configs", "tests", "*.tests", "*.tests.*", "tests.*")),
    python_requires=REQUIRES_PYTHON,
    install_requires=[
        "pyxb",  # need to install before ismrmrd
        "h5py",
        "matplotlib",
        "numpy",
        "tensorboard",
        "fvcore",
        "mridata",
        "scikit-image>=0.18.2",
        "sigpy>=0.1.17",
        "ismrmrd",
        "pandas",
        "silx",
        "tqdm",
        "omegaconf",
        "torchmetrics>=0.5.1",
        "iopath",
        "packaging",
    ],
    extras_require={
        "dev": [
            # Formatting
            "coverage",
            "flake8",
            "isort",
            "black==21.10b0",
            "flake8-bugbear",
            "flake8-comprehensions",
            "pre-commit>=2.9.3",
            # Testing
            "medpy",
            "pooch",
            "gdown",
            # Documentation
            "sphinx",
            "sphinxcontrib-bibtex",
            "sphinx-rtd-theme",
            "m2r2",
        ],
        "benchmarking": ["medpy"],
        "deployment": ["gdown"],
        "docs": ["sphinx", "sphinxcontrib.bibtex", "sphinx-rtd-theme", "m2r2"],
    },
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



from typing import Any, List, Mapping, Sequence, Union

import torch

from meddlr.config import CfgNode
from meddlr.transforms.build import TRANSFORM_REGISTRY, build_transforms
from meddlr.transforms.mixins import DeviceMixin
from meddlr.transforms.tf_scheduler import SchedulableMixin
from meddlr.transforms.transform import NoOpTransform, Transform
from meddlr.transforms.transform_gen import TransformGen

__all__ = ["RandomTransformChoice"]

_TRANSFORM_OR_GEN = Union[Transform, TransformGen]


@TRANSFORM_REGISTRY.register()
class RandomTransformChoice(TransformGen):
    def __init__(
        self, tfms_or_gens: Sequence[Union[Transform, TransformGen]], tfm_ps="uniform", p=0.0
    ) -> None:
        N = len(tfms_or_gens)
        if tfm_ps == "uniform":
            tfm_ps = torch.ones(N) / N
        else:
            tfm_ps = torch.as_tensor(tfm_ps)
        assert torch.allclose(torch.sum(tfm_ps), torch.as_tensor(1.0))
        self.tfm_ps = tfm_ps
        self.tfms_or_gens = tfms_or_gens

        super().__init__(p=p)

    def get_transform(self, input=None) -> Union[_TRANSFORM_OR_GEN, Sequence[_TRANSFORM_OR_GEN]]:
        params = self._get_param_values(use_schedulers=True)
        p = params["p"]
        if self._rand() >= p:
            return NoOpTransform()

        return self.tfms_or_gens[self._rand_choice(probs=self.tfm_ps)]

    def schedulers(self):
        tfms: List[SchedulableMixin] = self._get_tfm_by_type(SchedulableMixin)
        schedulers = list(self._schedulers)
        schedulers.extend([sch for tfm in tfms for sch in tfm.schedulers()])
        return schedulers

    def seed(self, value: int):
        self._generator = torch.Generator(device=self._device).manual_seed(value)
        tfms: List[TransformGen] = self._get_tfm_by_type(TransformGen)
        for t in tfms:
            t.seed(value)
        return self

    def to(self, device):
        super().to(device)
        tfms: List[DeviceMixin] = self._get_tfm_by_type(DeviceMixin)
        for t in tfms:
            t.to(device)
        return self

    def _get_tfm_by_type(self, klass):
        tfms = []
        for tfm in self.tfms_or_gens:
            if isinstance(tfm, (list, tuple)):
                tfms.extend(t for t in tfm if isinstance(t, klass))
            elif isinstance(tfm, klass):
                tfms.append(tfm)
        return tfms

    def __repr__(self):
        classname = type(self).__name__
        argstr = ",\n  ".join(
            "(p={:0.2f}) {}".format(p, repr(t)) for t, p in zip(self.tfms_or_gens, self.tfm_ps)
        )
        return "{}(\n  {}\n)".format(classname, argstr)

    @classmethod
    def from_dict(cls, cfg: CfgNode, init_kwargs: Mapping[str, Any], **kwargs):
        init_kwargs = init_kwargs.copy()
        tfms_or_gens = []
        for tfm_cfg in init_kwargs.pop("tfms_or_gens"):
            tfms_or_gens.append(build_transforms(cfg, tfm_cfg, **kwargs))

        return cls(tfms_or_gens, **init_kwargs)



"""Implementation of different dataset samplers."""
from typing import Sequence

import numpy as np
import torch
from torch.utils.data import Sampler

__all__ = ["AlternatingSampler"]


class AlternatingSampler(Sampler):
    """Sampler that alternates between supervised and unsupervised examples.

    In many semi-supervised scenarios, we would like to switch between using
    supervised and unsupervised examples. This sampler facilitates returning
    samples in this alternating pattern.

    Note:
        Currently, periods :math:`T_s` and :math:`T_us` must be perfect divisors
        of the number of supervised_idxs and unsupervised_idxs.
    """

    def __init__(
        self, dataset, T_s: int, T_us: int, indices: Sequence[int] = None, seed: int = None
    ):
        """
        Args:
            dataset: dataset to sample from. Must have methods
                `get_supervised_idxs()` and `get_unsupervised_idxs()`
            T_s (int): Period for returning supervised examples
            T_us (int): Period for returning unsupervised examples
            indices (Sequence[int], optional): Indices in the dataset to consider.
                If specified, only these indices will be iterated over.
        """
        super().__init__(dataset)

        supervised_idxs = dataset.get_supervised_idxs()
        unsupervised_idxs = dataset.get_unsupervised_idxs()
        if indices is not None:
            supervised_idxs = [x for x in supervised_idxs if x in indices]
            unsupervised_idxs = [x for x in unsupervised_idxs if x in indices]
        self._supervised_idxs = torch.tensor(supervised_idxs)
        self._unsupervised_idxs = torch.tensor(unsupervised_idxs)

        if len(self._unsupervised_idxs) == 0 or len(self._supervised_idxs) == 0:
            raise ValueError(
                "AlternatingSampler can only be used for "
                "semi-supervised training. "
                "Dataset must have both supervised/unsupervised examples"
            )

        if len(self._supervised_idxs) % T_s != 0:
            raise ValueError(
                "Period T_s must be a perfect divisor of "
                "the number of supervised indices. "
                "Got {} supervised indices, and T_s={}".format(len(self._supervised_idxs), T_s)
            )
        if len(self._unsupervised_idxs) % T_s != 0:
            raise ValueError(
                "Period T_us must be a perfect divisor of "
                "the number of unsupervised indices. "
                "Got {} unsupervised indices, and T_us={}".format(
                    len(self._unsupervised_idxs), T_us
                )
            )

        self.T_s = T_s
        self.T_us = T_us

        if seed is None:
            rng = torch.Generator().set_state(torch.random.get_rng_state())
        else:
            rng = torch.Generator().manual_seed(seed)
        self._rng = rng

        # The least common multiple between the number of
        # supervised block and number of unsupervised blocks in single pass.
        # num_blocks = num_examples / num_period
        num_blocks_supervised = int(len(self._supervised_idxs) / T_s)
        num_blocks_unsupervised = int(len(self._unsupervised_idxs) / T_us)
        self._lcm = int(np.lcm(num_blocks_supervised, num_blocks_unsupervised))
        # The sampler is determined by how many blocks of each data type
        # (supervised/unsupervised) are returned.
        # Number of passes indicates how many times we have to go through
        # the dataset based on the least common multiple.
        num_passes_supervised = int(self._lcm / num_blocks_supervised)
        num_passes_unsupervised = int(self._lcm / num_blocks_unsupervised)
        self._num_samples_sup = num_passes_supervised * len(self._supervised_idxs)  # noqa
        self._num_samples_unsup = num_passes_unsupervised * len(self._unsupervised_idxs)  # noqa

    def get_indices(self, as_list=False):
        s_idxs = self._build_idx(self._supervised_idxs, self.T_s, self._num_samples_sup)
        us_idxs = self._build_idx(self._unsupervised_idxs, self.T_us, self._num_samples_unsup)

        # TODO: Add parameter to choose whether to start with supervised or
        # unsupervised data.
        idxs = torch.cat([s_idxs, us_idxs], dim=1).reshape(-1)
        if as_list:
            idxs = idxs.tolist()
        return idxs

    def __len__(self):
        return self._num_samples_sup + self._num_samples_unsup

    def __iter__(self):
        return iter(self.get_indices(as_list=True))

    def _build_idx(self, idxs: torch.Tensor, T: int, num_samples: int):
        assert num_samples % len(idxs) == 0
        num_passes = int(num_samples / len(idxs))

        perm_idxs = torch.cat(
            [torch.randperm(len(idxs), generator=self._rng) for _ in range(num_passes)]
        )

        idxs = idxs[perm_idxs].reshape(-1, T)
        return idxs



from meddlr.transforms.base.mask import KspaceMaskTransform  # noqa
from meddlr.transforms.base.motion import MRIMotionTransform  # noqa
from meddlr.transforms.base.noise import NoiseTransform  # noqa
from meddlr.transforms.base.spatial import (  # noqa
    AffineTransform,
    FlipTransform,
    Rot90Transform,
    TranslationTransform,
)



import inspect
import multiprocessing as mp
from typing import Any, Dict, List, Mapping, Union

import numpy as np
from fvcore.common.registry import Registry

from meddlr.config import CfgNode
from meddlr.transforms.tf_scheduler import TFScheduler, WarmupMultiStepTF, WarmupTF
from meddlr.transforms.transform import Transform
from meddlr.transforms.transform_gen import TransformGen

TRANSFORM_REGISTRY = Registry("TRANSFORM_FUNCS")  # noqa F401 isort:skip
TRANSFORM_REGISTRY.__doc__ = """
Registry for transform functions.

The registered object will be called with `obj(cfg)`
and expected to return a `nn.Module` object.
"""


def build_transforms(
    cfg: CfgNode, tfm_cfgs: List[Mapping[str, Any]], seed=None, **kwargs
) -> List[Union[Transform, TransformGen]]:
    # This should only be used for training right now.
    is_single = isinstance(tfm_cfgs, Mapping)
    if is_single:
        tfm_cfgs = [tfm_cfgs]

    tfms = [_build_transform(cfg, tfm_cfg, **kwargs) for tfm_cfg in tfm_cfgs]

    if seed is not None:
        seed_tfm_gens(tfms, seed=seed)

    if is_single:
        tfms = tfms[0]

    return tfms


def _build_transform(cfg: CfgNode, tfm_cfg: Dict[str, Any], **kwargs):
    tfm_cfg = tfm_cfg.copy()
    name: str = tfm_cfg.pop("name")
    schedulers_cfgs = tfm_cfg.pop("schedulers", tfm_cfg.pop("scheduler", None))
    init_args: Mapping[str, Any] = tfm_cfg

    klass = TRANSFORM_REGISTRY.get(name)
    parameters = inspect.signature(klass.__init__).parameters
    init_args.update(**{k: v for k, v in kwargs.items() if k in parameters})

    if hasattr(klass, "from_dict"):
        tfm = klass.from_dict(cfg, init_args, **kwargs)
    else:
        try:
            tfm = klass(**init_args)
        except TypeError as e:
            raise TypeError(f"Failed to initialize {klass.__name__} - {e}")
        except RecursionError as e:
            raise RecursionError(f"Failed to initialize {klass.__name__} - {e}")
    if isinstance(tfm, Transform) or schedulers_cfgs is None:
        return tfm

    tfm_gen: TransformGen = tfm
    if isinstance(schedulers_cfgs, Mapping):
        schedulers_cfgs = [schedulers_cfgs]
    schedulers = [build_scheduler(cfg, scfg, tfm_gen) for scfg in schedulers_cfgs]
    tfm_gen.register_schedulers(schedulers)
    return tfm_gen


def build_scheduler(
    cfg: CfgNode, scheduler_cfg: Dict[str, Any], tfm_gen: TransformGen
) -> Dict[str, TFScheduler]:
    scheduler_cfg = scheduler_cfg.copy()
    name: str = scheduler_cfg.pop("name")
    params: List[str] = scheduler_cfg.pop("params", None)

    if not name:
        return None
    elif name == "WarmupMultiStepLR":
        return WarmupMultiStepTF(tfm_gen, params=params, **scheduler_cfg)
    elif name == "WarmupTF":
        return WarmupTF(tfm_gen, params=params, **scheduler_cfg)
    elif name == "WarmupStepTF":
        step = scheduler_cfg["warmup_milestones"]
        delay_iters = scheduler_cfg.pop("delay_iters", 0)
        if isinstance(step, int):
            step = (step,)
        if len(step) != 1:
            raise ValueError("step must have single value for WarmupStepTF")
        step = step[0]
        max_iter = scheduler_cfg.pop(
            "max_iter", scheduler_cfg.pop("max_iters", cfg.SOLVER.MAX_ITER)
        )
        num_steps = (max_iter - delay_iters) // step
        steps = tuple(step * x + delay_iters for x in range(1, 1 + num_steps))
        if delay_iters > 0:
            steps = (delay_iters,) + steps
        init_kwargs = dict(tfm=tfm_gen, warmup_milestones=steps, params=params)
        init_kwargs.update({k: v for k, v in scheduler_cfg.items() if k not in init_kwargs})
        return WarmupMultiStepTF(**init_kwargs)

    raise ValueError("Unknown LR scheduler: {}".format(name))


def seed_tfm_gens(tfms, seed):
    # Seed all transform generators with unique, but reproducible seeds.
    # Do not change the scaling constant (1e10).
    rng = np.random.RandomState(seed)
    if seed is not None:
        for t in tfms:
            if isinstance(t, TransformGen):
                t.seed(int(rng.rand() * 1e10))


def build_iter_func(batch_size, num_workers):
    def get_iter(step):
        # Worker id is 1-indexed by default.
        worker_id = int(mp.current_process().name.split("-")[1])
        curr_iter = (step * num_workers + worker_id - 1) // batch_size
        return curr_iter

    return get_iter



import logging

from fvcore.common.checkpoint import Checkpointer as _Checkpointer
from fvcore.common.checkpoint import _strip_prefix_if_present

logger = logging.getLogger(__name__)


class Checkpointer(_Checkpointer):
    """
    Extension of fvcore's Checkpointer class that handles backwards compatibility
    and systematic state_dict unpacking.

    To use the full functionality of this class, all saved models should follow
    the convention that wrapped models begin with the prefix ``'model.'``.
    See :cls:`ss_recon.modeling.meta_arch.GeneralizedUnrolledCNN`.
    """

    def __init__(self, model, save_dir="", *, save_to_disk=None, **checkpointables):
        super().__init__(model, save_dir, save_to_disk=True, **checkpointables)

    def _load_file(self, filename):
        loaded = super()._load_file(filename)  # load native pth checkpoint
        if "model" not in loaded:
            loaded = {"model": loaded}
        return loaded

    def _load_model(self, checkpoint):
        model_state_dict = checkpoint["model"]
        incompatible = super()._load_model(checkpoint)
        if not _has_incompatible_keys(incompatible):
            return

        # Load models that are wrapped in an nn.Module.
        # Following convention, these wrapped models begin with the "model." prefix.
        logger.warn("Attempting to load wrapped model from model state dict.")
        _strip_prefix_if_present(model_state_dict, "module.")
        _strip_prefix_if_present(model_state_dict, "model.")
        incompatible = super()._load_model({"model": model_state_dict})
        if _has_incompatible_keys(incompatible):
            raise ValueError("Found incompatible keys or shapes:\n{}".format(incompatible))


def _has_incompatible_keys(incompatible):
    return (
        bool(incompatible.missing_keys)
        | bool(incompatible.unexpected_keys)
        | bool(incompatible.incorrect_shapes)
    )



import inspect
import os
from typing import Sequence

import numba as nb
import numpy as np
import sigpy.mri
import torch
from fvcore.common.registry import Registry

MASK_FUNC_REGISTRY = Registry("MASK_FUNC")
MASK_FUNC_REGISTRY.__doc__ = """
Registry for mask functions, which create undersampling masks of a specified
shape.
"""


def build_mask_func(cfg, **kwargs):
    name = cfg.UNDERSAMPLE.NAME
    accelerations = cfg.UNDERSAMPLE.ACCELERATIONS
    calibration_size = cfg.UNDERSAMPLE.CALIBRATION_SIZE
    center_fractions = cfg.UNDERSAMPLE.CENTER_FRACTIONS

    klass = MASK_FUNC_REGISTRY.get(name)
    parameters = inspect.signature(klass).parameters

    # Optional args
    kwargs = kwargs.copy()
    mapping = {"max_attempts": cfg.UNDERSAMPLE.MAX_ATTEMPTS}
    for param, value in mapping.items():
        if param in parameters:
            kwargs[param] = value

    return klass(accelerations, calibration_size, center_fractions, **kwargs)


class MaskFunc:
    """Abstract MaskFunc class for creating undersampling masks of a specified
    shape.

    Adapted from Facebook fastMRI.
    """

    def __init__(self, accelerations):
        """
        Args:
            accelerations (List[int]): Range of acceleration rates to simulate.
        """
        self.accelerations = accelerations
        self.rng = np.random.RandomState()

    def choose_acceleration(self):
        """Chooses a random acceleration rate given a range.

        If self.accelerations is a constant, it will be returned

        """
        if not isinstance(self.accelerations, Sequence):
            return self.accelerations
        elif len(self.accelerations) == 1:
            return self.accelerations[0]
        accel_range = self.accelerations[1] - self.accelerations[0]
        acceleration = self.accelerations[0] + accel_range * self.rng.rand()
        return acceleration


class CacheableMaskMixin:
    def get_filename(self):
        raise NotImplementedError


@MASK_FUNC_REGISTRY.register()
class RandomMaskFunc(MaskFunc):
    """
    RandomMaskFunc creates a 2D uniformly random undersampling mask.
    """

    def __init__(self, accelerations, calib_size, center_fractions=None):
        if center_fractions:
            raise ValueError(f"center_fractions not yet supported for class {type(self)}.")
        super().__init__(accelerations)
        self.calib_size = calib_size

    def __call__(self, out_shape, seed=None, acceleration=None):
        # Design parameters for mask
        nky = out_shape[1]
        nkz = out_shape[2]

        if not acceleration:
            acceleration = self.choose_acceleration()
        prob = 1.0 / acceleration

        # Generate undersampling mask.
        rand_kwargs = {"dtype": torch.float32}
        if seed is not None:
            rand_kwargs["generator"] = torch.Generator().manual_seed(seed)

        mask = torch.rand([nky, nkz], **rand_kwargs)
        mask = torch.where(mask < prob, torch.Tensor([1]), torch.Tensor([0]))

        # Add calibration region
        calib = [self.calib_size, self.calib_size]
        mask[
            int(nky / 2 - calib[-2] / 2) : int(nky / 2 + calib[-2] / 2),
            int(nkz / 2 - calib[-1] / 2) : int(nkz / 2 + calib[-1] / 2),
        ] = torch.Tensor([1])

        return mask.reshape(out_shape)


@MASK_FUNC_REGISTRY.register()
class PoissonDiskMaskFunc(CacheableMaskMixin, MaskFunc):
    """
    PoissonDiskMaskFunc creates a 2D Poisson disk undersampling mask.
    """

    def __init__(
        self,
        accelerations,
        calib_size,
        center_fractions=None,
        max_attempts=30,
        crop_corner=True,
        module="internal",
    ):
        if center_fractions:
            raise ValueError(f"center_fractions not yet supported for class {type(self)}.")
        if module not in ("internal", "sigpy"):
            raise ValueError("`module` must be one of ('internal', 'sigpy')")
        super().__init__(accelerations)
        if isinstance(calib_size, int):
            calib_size = (calib_size, calib_size)
        self.calib_size = calib_size
        self.max_attempts = max_attempts
        self.crop_corner = crop_corner
        self.module = module

    def __call__(self, out_shape, seed=None, acceleration=None):
        # Design parameters for mask
        nky = out_shape[1]
        nkz = out_shape[2]
        if not acceleration:
            acceleration = self.choose_acceleration()

        # From empirical results, larger dimension should be first
        # for optimal speed.
        if nky < nkz:
            shape = (nkz, nky)
            transpose = True
        else:
            shape = (nky, nkz)
            transpose = False

        # Issue #2: Due to some optimization reasons, using the internal
        # poisson disc module has been faster than using sigpy's
        # default one. In many cases, the call to sigpy hangs.
        module = self.module
        if module == "internal":
            mask = poisson(
                shape,
                acceleration,
                calib=self.calib_size,
                dtype=np.float32,
                seed=seed,
                K=self.max_attempts,
                crop_corner=self.crop_corner,
            )
        elif module == "sigpy":
            mask = sigpy.mri.poisson(
                shape,
                acceleration,
                calib=self.calib_size,
                dtype=np.float32,
                seed=seed,
                max_attempts=self.max_attempts,
                crop_corner=self.crop_corner,
            )
        if transpose:
            mask = mask.transpose()

        # Reshape the mask
        mask = torch.from_numpy(mask.reshape(out_shape))

        return mask

    def _get_args(self):
        return {
            "accelerations": self.accelerations,
            "calib_size": self.calib_size,
            "max_attempts": self.max_attempts,
            "crop_corner": self.crop_corner,
        }

    def get_str_name(self):
        args = self._get_args()
        return f"{type(self).__name__}-" + "-".join(f"{k}={v}" for k, v in args.items())

    def __str__(self) -> str:
        args = self._get_args()
        args_str = "\n\t" + "\n\t".join(f"{k}={v}" for k, v in args.items()) + "\n\t"
        return f"{type(self)}({args_str})"


@MASK_FUNC_REGISTRY.register()
class RandomMaskFunc1D(MaskFunc):
    """
    RandomMaskFunc creates a sub-sampling mask of a given shape.
    The mask selects a subset of columns from the input k-space data. If the k-space data has N
    columns, the mask picks out:
        1. N_low_freqs = (N * center_fraction) columns in the center corresponding to
           low-frequencies
        2. The other columns are selected uniformly at random with a probability equal to:
           prob = (N / acceleration - N_low_freqs) / (N - N_low_freqs).
    This ensures that the expected number of columns selected is equal to (N / acceleration)
    It is possible to use multiple center_fractions and accelerations, in which case one possible
    (center_fraction, acceleration) is chosen uniformly at random each time the RandomMaskFunc
    object is called.

    For example, if accelerations = [4, 8] and center_fractions = [0.08, 0.04], then there
    is a 50% probability that 4-fold acceleration with 8% center fraction is selected and a 50%
    probability that 8-fold acceleration with 4% center fraction is selected.

    Adapted from https://github.com/facebookresearch/fastMRI/blob/master/common/subsample.py
    """

    def __init__(self, accelerations, calib_size=None, center_fractions=None):
        """
        Args:
            center_fractions (List[float]): Fraction of low-frequency columns to be retained.
                If multiple values are provided, then one of these numbers is chosen uniformly
                each time.
            accelerations (List[int]): Amount of under-sampling. This should have the same length
                as center_fractions. If multiple values are provided, then one of these is chosen
                uniformly each time. An acceleration of 4 retains 25% of the columns, but they may
                not be spaced evenly.
            calib_size (List[int]): Calibration size for scans.
        """
        if not calib_size and not center_fractions:
            raise ValueError("Either calib_size or center_fractions must be specified.")
        if calib_size and center_fractions:
            raise ValueError("Only one of calib_size or center_fractions can be specified")

        self.center_fractions = center_fractions
        self.calib_size = calib_size
        self.accelerations = accelerations

    def __call__(self, shape, seed=None, acceleration=None):
        """
        Args:
            shape (iterable[int]): The shape of the mask to be created. The shape should have
                at least 3 dimensions. Samples are drawn along the second last dimension.
            seed (int, optional): Seed for the random number generator. Setting the seed
                ensures the same mask is generated each time for the same shape.
        Returns:
            torch.Tensor: A mask of the specified shape.
        """
        if len(shape) < 3:
            raise ValueError("Shape should have 3 or more dimensions")

        if seed is not None:
            np_state = np.random.get_state()

        num_rows = shape[1]
        num_cols = shape[2]
        if self.center_fractions:
            if isinstance(self.center_fractions, Sequence):
                choice = np.random.randint(0, len(self.center_fractions))
                center_fraction = self.center_fractions[choice]
            else:
                center_fraction = self.center_fractions
        else:
            center_fraction = self.calib_size / num_cols
        if acceleration is None:
            acceleration = self.choose_acceleration()

        # Create the mask
        num_low_freqs = int(round(num_cols * center_fraction))
        prob = (num_cols / acceleration - num_low_freqs) / (num_cols - num_low_freqs)
        mask = np.random.uniform(size=num_cols) < prob
        pad = (num_cols - num_low_freqs + 1) // 2
        mask[pad : pad + num_low_freqs] = True

        # Reshape the mask
        mask_shape = [1 for _ in shape]
        mask_shape[2] = num_cols
        mask = mask.reshape(*mask_shape).astype(np.float32)
        mask = np.concatenate([mask] * num_rows, axis=1)
        mask = torch.from_numpy(mask)

        if seed is not None:
            np.random.set_state(np_state)

        return mask


class MaskLoader(MaskFunc):
    """Loads masks from predefined file format instead of computing on the fly."""

    def __init__(self, accelerations, masks_path, mask_type: str = "poisson", mode="train"):
        assert isinstance(accelerations, (int, float)) or len(accelerations) == 1
        assert mode in ["train", "eval"]
        if isinstance(accelerations, (int, float)):
            accelerations = (accelerations,)
        super().__init__(accelerations)

        accel = float(self.accelerations[0])
        self.train_masks = None
        self.eval_data = torch.load(os.path.join(masks_path, f"{mask_type}_{accel}x_eval.pt"))
        if mode == "train":
            self.train_masks = np.load(os.path.join(masks_path, f"{mask_type}_{accel}x.npy"))

    def __call__(self, out_shape, seed=None, acceleration=None):
        if acceleration is not None and acceleration not in self.accelerations:
            raise RuntimeError(
                "MaskLoader.__call__ does not currently support ``acceleration`` argument"
            )

        if seed is None:
            # Randomly select from the masks we have
            idx = np.random.choice(len(self.train_masks))
            mask = self.train_masks[idx]
        else:
            data = self.eval_data
            masks = self.eval_data["masks"]
            mask = masks[data["seeds"].index(seed)]

        mask = mask.reshape(out_shape)
        return torch.from_numpy(mask)


# ================================================================ #
# Adapted from sigpy.
# Duplicated because of https://github.com/mikgroup/sigpy/issues/54
# TODO: Remove once https://github.com/mikgroup/sigpy/issues/54 is
# solved and added to release.
# ================================================================ #
def poisson(
    img_shape,
    accel,
    K=30,
    calib=(0, 0),
    dtype=np.complex,
    crop_corner=True,
    return_density=False,
    seed=0,
):
    """Generate Poisson-disc sampling pattern

    Args:
        img_shape (tuple of ints): length-2 image shape.
        accel (float): Target acceleration factor. Greater than 1.
        K (float): maximum number of samples to reject.
        calib (tuple of ints): length-2 calibration shape.
        dtype (Dtype): data type.
        crop_corner (bool): Toggle whether to crop sampling corners.
        return_density (bool): Toggle whether to return sampling density.
        seed (int): Random seed.

    Returns:
        array: Poisson-disc sampling mask.

    References:
        Bridson, Robert. "Fast Poisson disk sampling in arbitrary dimensions."
        SIGGRAPH sketches. 2007.

    """
    y, x = np.mgrid[: img_shape[-2], : img_shape[-1]]
    x = np.maximum(abs(x - img_shape[-1] / 2) - calib[-1] / 2, 0)
    x /= x.max()
    y = np.maximum(abs(y - img_shape[-2] / 2) - calib[-2] / 2, 0)
    y /= y.max()
    r = np.sqrt(x ** 2 + y ** 2)

    slope_max = 40
    slope_min = 0
    if seed is not None:
        rand_state = np.random.get_state()
    else:
        seed = -1  # numba does not play nicely with None types
    while slope_min < slope_max:
        slope = (slope_max + slope_min) / 2.0
        R = 1.0 + r * slope
        mask = _poisson(img_shape[-1], img_shape[-2], K, R, calib, seed)
        if crop_corner:
            mask *= r < 1

        est_accel = img_shape[-1] * img_shape[-2] / np.sum(mask[:])

        if abs(est_accel - accel) < 0.1:
            break
        if est_accel < accel:
            slope_min = slope
        else:
            slope_max = slope

    if seed is not None and seed > 0:
        np.random.set_state(rand_state)
    mask = mask.reshape(img_shape).astype(dtype)
    if return_density:
        return mask, r
    else:
        return mask


@nb.jit(nopython=True, cache=True)  # pragma: no cover
def _poisson(nx, ny, K, R, calib, seed=None):

    mask = np.zeros((ny, nx))
    f = ny / nx

    if seed is not None and seed > 0:
        np.random.seed(int(seed))

    pxs = np.empty(nx * ny, np.int32)
    pys = np.empty(nx * ny, np.int32)
    pxs[0] = np.random.randint(0, nx)
    pys[0] = np.random.randint(0, ny)
    m = 1
    while m > 0:

        i = np.random.randint(0, m)
        px = pxs[i]
        py = pys[i]
        rad = R[py, px]

        # Attempt to generate point
        done = False
        k = 0
        while not done and k < K:

            # Generate point randomly from R and 2R
            rd = rad * (np.random.random() * 3 + 1) ** 0.5
            t = 2 * np.pi * np.random.random()
            qx = px + rd * np.cos(t)
            qy = py + rd * f * np.sin(t)

            # Reject if outside grid or close to other points
            if qx >= 0 and qx < nx and qy >= 0 and qy < ny:

                startx = max(int(qx - rad), 0)
                endx = min(int(qx + rad + 1), nx)
                starty = max(int(qy - rad * f), 0)
                endy = min(int(qy + rad * f + 1), ny)

                done = True
                for x in range(startx, endx):
                    for y in range(starty, endy):
                        if mask[y, x] == 1 and (
                            ((qx - x) / R[y, x]) ** 2 + ((qy - y) / (R[y, x] * f)) ** 2 < 1
                        ):
                            done = False
                            break

                    if not done:
                        break

            k += 1

        # Add point if done else remove active
        if done:
            pxs[m] = qx
            pys[m] = qy
            mask[int(qy), int(qx)] = 1
            m += 1
        else:
            pxs[i] = pxs[m - 1]
            pys[i] = pys[m - 1]
            m -= 1

    # Add calibration region
    mask[
        int(ny / 2 - calib[-2] / 2) : int(ny / 2 + calib[-2] / 2),
        int(nx / 2 - calib[-1] / 2) : int(nx / 2 + calib[-1] / 2),
    ] = 1

    return mask



"""Default setup in meddlr.

This file contains components with some default boilerplate logic user may need
in training / testing. They will not work for everyone, but many users may find
them useful.

The behavior of functions/classes in this file is subject to change,
since they are meant to represent the "common default behavior" people need in
their projects.
"""

import argparse
import os
import re
import warnings
from typing import Mapping, Sequence

import torch

from meddlr.config.config import CfgNode
from meddlr.utils import env
from meddlr.utils.collect_env import collect_env_info
from meddlr.utils.env import get_available_gpus, seed_all_rng
from meddlr.utils.logger import setup_logger

__all__ = ["default_argument_parser", "default_setup"]

_PATH_MANAGER = env.get_path_manager()


def default_argument_parser() -> argparse.ArgumentParser:
    """Create a parser with some common arguments.

    Returns:
        argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(description="SS-Recon Training")
    parser.add_argument("--config-file", default="", metavar="FILE", help="path to config file")
    parser.add_argument(
        "--resume",
        action="store_true",
        help="whether to attempt to resume from the checkpoint directory",
    )
    parser.add_argument(
        "--restart-iter",
        action="store_true",
        help="restart iteration count when loading checkpointed weights",
    )
    parser.add_argument("--eval-only", action="store_true", help="perform evaluation only")
    parser.add_argument(
        "--num-gpus", type=int, default=1, help="number of gpus. overrided by --devices"
    )
    parser.add_argument("--devices", type=int, nargs="*", default=None)
    parser.add_argument("--debug", action="store_true", help="use debug mode")
    parser.add_argument(
        "--reproducible", "--repro", action="store_true", help="activate reproducible mode"
    )

    parser.add_argument(
        "opts",
        help="Modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )
    return parser


def default_setup(cfg, args, save_cfg: bool = True):
    """
    Perform some basic common setups at the beginning of a job, including:

    1. Set up the ss_recon logger
    2. Log basic information about environment, cmdline arguments, and config
    3. Backup the config to the output directory
    4. Enables debug mode if ``args.debug==True``
    5. Enables reproducible model if ``args.reproducible==True``

    Args:
        cfg (CfgNode): the full config to be used
        args (argparse.NameSpace): the command line arguments to be logged
        save_cfg (bool, optional): If `True`, writes config to `cfg.OUTPUT_DIR`.

    Note:
        Project-specific environment variables are modified by this function.
        ``cfg`` is also modified in-place.
    """
    is_repro_mode = (
        env.is_repro() if env.is_repro() else (hasattr(args, "reproducible") and args.reproducible)
    )
    eval_only = hasattr(args, "eval_only") and args.eval_only

    # Update config parameters before saving.
    cfg.defrost()
    cfg.OUTPUT_DIR = _PATH_MANAGER.get_local_path(cfg.OUTPUT_DIR)
    if is_repro_mode:
        init_reproducible_mode(cfg, eval_only)
    cfg.freeze()

    output_dir = cfg.OUTPUT_DIR
    if output_dir:
        _PATH_MANAGER.mkdirs(output_dir)

    setup_logger(output_dir, name="fvcore")
    logger = setup_logger(output_dir)

    if args.debug:
        os.environ["MEDDLR_DEBUG"] = "True"
        logger.info("Running in debug mode")
    if is_repro_mode:
        os.environ["MEDDLR_REPRO"] = "True"
        logger.info("Running in reproducible mode")

    logger.info("Environment info:\n" + collect_env_info())

    logger.info("Command line arguments: " + str(args))
    if hasattr(args, "config_file") and args.config_file != "":
        logger.info(
            "Contents of args.config_file={}:\n{}".format(
                args.config_file, _PATH_MANAGER.open(args.config_file, "r").read()
            )
        )

    if "CUDA_VISIBLE_DEVICES" not in os.environ:
        if args.devices:
            gpus = args.devices
            if not isinstance(gpus, Sequence):
                gpus = [gpus]
        else:
            gpus = get_available_gpus(args.num_gpus)

        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join([str(x) for x in gpus])

        # TODO: Remove this and find a better way to launch the script
        # with the desired gpus.
        if gpus[0] >= 0:
            torch.cuda.set_device(gpus[0])

    logger.info("Running with full config:\n{}".format(cfg))
    if output_dir and save_cfg:
        # Note: some of our scripts may expect the existence of
        # config.yaml in output directory
        path = os.path.join(output_dir, "config.yaml")
        with _PATH_MANAGER.open(path, "w") as f:
            f.write(cfg.dump())
        logger.info("Full config saved to {}".format(os.path.abspath(path)))

    # make sure each worker has a different, yet deterministic seed if specified
    seed_all_rng(None if cfg.SEED < 0 else cfg.SEED)

    # cudnn benchmark has large overhead.
    # It shouldn't be used considering the small size of
    # typical validation set.
    if not eval_only:
        torch.backends.cudnn.benchmark = cfg.CUDNN_BENCHMARK


def find_wandb_exp_id(cfg):
    """Helper function to find W&B experiment id.

    Args:
        cfg (CfgNode): The config.

    Returns:
        exp_id (Optional[str]): Returns `None` if experiment id
            cannot be determined.
    """
    filepath = os.path.join(cfg.OUTPUT_DIR, "wandb_id")
    exp_id = None
    # Option 1 (preferred): Check if wandb experiment file exists.
    if os.path.isfile(filepath):
        with open(filepath, "r") as f:
            exp_id = f.read()
        return exp_id

    # Option 2: Navigate folder structure to find wandb relevant folders.
    # Note this may not be robust as `wandb` may change their folder structure.
    base_dir = os.path.join(cfg.OUTPUT_DIR, "wandb", "latest-run")
    if os.path.isdir(base_dir):
        run_files = [x for x in os.listdir(base_dir) if os.path.isfile(os.path.join(base_dir, x))]
        assert len(run_files) == 1, "Found multiple ({}) W&B run files:\n\t{}".format(
            len(run_files), "\n\t".join(run_files)
        )
        exp_id = os.path.splitext(run_files[0])[0].split("-")[1]
    return exp_id


def init_wandb_run(
    cfg,
    exp_id=None,
    resume=False,
    project=None,
    entity=None,
    job_type: str = "training",
    sync_tensorboard: bool = None,
    use_api: bool = False,
):
    """Helper function to initialize Weights & Biases based on project file structure.

    TODO:
        1. Configure `sync_tensorboard` with pytorch lightning once
            new W&B version comes out. (escalated to W&B)
        2. Configure `resume` functionality to log to the same run
            without overwriting. (escalated to W&B)
        3. Configure `is_eval` to use `wandb.init` (related to 2)

    """
    import wandb

    if sync_tensorboard is not None:
        os.environ["WANDB_SYNC_TENSORBOARD"] = "true" if sync_tensorboard else "false"
    sync_tensorboard = os.environ.get("WANDB_SYNC_TENSORBOARD", False)

    # Find last run if `exp_id` not specified.
    is_eval = job_type.lower() in ("eval", "evaluation")
    if (resume or is_eval) and not exp_id:
        exp_id = find_wandb_exp_id(cfg)

    # If evaluation, do not run wandb.init. Just return run.
    if use_api:
        if not exp_id:
            raise ValueError("No experiment id found.")
        api = wandb.Api()
        return api.run(f"{project}/{entity}/{exp_id}")

    if project is None:
        project = cfg.DESCRIPTION.PROJECT_NAME
    else:
        warnings.warn(
            "Setting project name with `project` is deprecated. "
            "Use DESCRIPTION.PROJECT_NAME in config instead.",
            DeprecationWarning,
        )
    if entity is None:
        entity = cfg.DESCRIPTION.ENTITY_NAME
    else:
        warnings.warn(
            "Setting entity name with `entity` is deprecated. "
            "Use DESCRIPTION.ENTITY_NAME in config instead.",
            DeprecationWarning,
        )

    # Keyword args to share between resumed and new runs.
    wandb_kwargs = dict(
        config=cfg,
        project=project,
        entity=entity,
        sync_tensorboard=sync_tensorboard,  # TODO: Change when W&B bug fixed
        job_type=job_type,
        dir=cfg.OUTPUT_DIR,
    )

    # Resume run and return.
    if exp_id:
        assert False, "We should not reach this until TODOs 1-3 are resolved."

    # Create new run.
    exp_name = cfg.DESCRIPTION.EXP_NAME
    if not exp_name:
        warnings.warn("DESCRIPTION.EXP_NAME not specified. Defaulting to basename...")
        exp_name = os.path.basename(cfg.OUTPUT_DIR)

    # Write the wandb id to a file.
    exp_id = wandb.util.generate_id()
    with open(os.path.join(cfg.OUTPUT_DIR, "wandb_id"), "w") as f:
        f.write(exp_id)

    wandb.init(
        id=exp_id,
        name=exp_name,
        tags=cfg.DESCRIPTION.TAGS,
        notes=cfg.DESCRIPTION.BRIEF,
        **wandb_kwargs,
    )
    return wandb.run


def init_reproducible_mode(cfg: CfgNode, eval_only: bool):
    """Activates reproducible mode and sets appropriate config paraemters.

    This method does the following:
        1. Sets environment variable indiciating project is in reproducible mode.
        2. Sets all seeds in the ``cfg`` if they are not initialized.
        3. Enables cuda benchmarking ``eval_only=False`` and ``cfg.CUDNN_BENCHMARK=True``.
        4. Sets ``torch.backends.cudann.deterministic=True``.

    Seed fields in ``cfg`` are indicated by keys that end with ``"SEED"``
    and whose corresponding value is an integer. Below are some examples of
    fields that would match as a seed field:

        * cfg.SEED = -1
        * cfg.XX.YY.SEED = -1
        * cfg.A_SEED = -1
        * cfg.SEED_VAL = -1  # this would not match, does not end with ``'SEED'``.
        * cfg.SEED = "alpha"  # this would not match, value is not an int.

    Args:
        cfg (CfgNode): The full config. This will be modified in place.
        eval_only (bool): If ``True``, initialize reproducible
            mode in an evaluation only setting.

    Note:
        This method should typically be called through :func:`default_setup`.
    """
    os.environ["MEDDLR_REPRO"] = "True"

    orig_cfg = cfg.clone()
    cfg.defrost()

    # Set all seeds in the config if they are not set.
    seed = seed_all_rng(None if cfg.SEED < 0 else cfg.SEED)
    _set_all_seeds(cfg, seed)

    # Turn off cuda benchmarking if this run is not only for evaluation.
    # If eval_only, default to current config value.
    cfg.CUDNN_BENCHMARK = False if not eval_only else orig_cfg.CUDNN_BENCHMARK
    torch.backends.cudnn.benchmark = cfg.CUDNN_BENCHMARK

    # Turn on deterministic mode.
    torch.backends.cudnn.deterministic = True

    cfg.freeze()


def _set_all_seeds(cfg, seed_val):
    for key, value in cfg.items():
        if re.match("^.*SEED$", key) and isinstance(value, int) and value < 0:
            cfg.__setattr__(key, seed_val)
        if isinstance(value, Mapping):
            _set_all_seeds(value, seed_val)



import itertools
import os
import pathlib
from typing import Any, Dict, List, Sequence, Union

from meddlr.utils import env

from .config import CfgNode


def configure_params(
    params: Dict[str, Sequence], fixed: Dict[str, Any] = None, base_cfg: CfgNode = None
) -> List[Dict]:
    """Unroll parameters into a list of configurations.

    Args:
        params (Dict[Sequence]): Config keys and potential values they can take on.
            These parameters will be mixed and matched to create configs.
            Order is ``{"param1": [valA, valB, ...], "param2": [val1, val2, ...]}``
        fixed (Dict[str, Any]): Fixed parameters to apply to every configuration.

    Returns:
        List[Dict] | List[CfgNode]: Different configurations. If ``base_cfg`` provided,
            the different configurations are merged into the config.
    """
    configs = itertools.product(*list(params.values()))
    configs = [{k: v for k, v in zip(params.keys(), cfg)} for cfg in configs]
    if fixed is not None:
        for c in configs:
            c.update(fixed)
    if base_cfg:
        configs = [base_cfg.clone().defrost().update(c).freeze() for c in configs]
    return configs


def stringify(cfg: Dict[str, Any]):
    """Convert param/value pairs into a command-line compatible string.

    Args:
        cfg (Dict[str, Any]): The configuration to stringify.

    Return:
        str: A command line compatible string.
    """
    cfg = {k: f"'{v}'" if isinstance(v, str) else v for k, v in cfg.items()}
    cfg = {k: _stringify_value(v) for k, v in cfg.items()}
    return " ".join(f'{k} "{v}"' for k, v in cfg.items())


def check_dependencies(
    cfg_file_or_lines, return_failed_deps: bool = False
) -> Union[bool, List[str]]:
    """Check that module dependencies are met for the config file.

    Dependencies are specified as comments in the config file starting
    with ``"# DEPENDENCY:"``.

    Args:
        cfg_file (str): The path to the config file.
        return_failed_deps (bool, optional): Whether to return the list of
            dependencies that are not met. Defaults to ``False``.

    Returns:
        bool | List[str]: If ``return_failed_deps=True``, returns the list of
        dependencies that are not met. Else returns boolean of whether all
        dependencies are met.
    """
    keyword = "# DEPENDENCY:"

    if isinstance(cfg_file_or_lines, (str, os.PathLike, pathlib.Path)):
        with open(cfg_file_or_lines, "r") as f:
            lines = f.readlines()
    else:
        lines = cfg_file_or_lines
    lines = [line.strip() for line in lines if keyword in line]
    dependencies = [
        dep.strip() for line in lines for dep in line.split(keyword)[-1].strip().split(";")
    ]

    missing_deps = []
    for dep in dependencies:
        if not env.is_package_installed(dep):
            missing_deps.append(dep)

    if return_failed_deps:
        return missing_deps
    else:
        return len(missing_deps) == 0


def _stringify_value(value, depth=0) -> str:
    """Convert value into command-line comaptible string.

    The string representations of certain types (e.g. list, tuple)
    are not command-line compatible. We have to add extra ``""``
    around the string for proper formatting. This function
    formats values of primitive types (numbers, strs, collections, dicts).

    Args:
        value (Number | str | Collection | Dict): The value to convert
            to a command-line compatible string.

    Returns:
        str: The stringified value.
    """
    if not isinstance(value, (set, tuple, list, dict)):
        return value
    if isinstance(value, dict):
        keys = list(value.keys())
        values = [value[k] for k in keys]
        keys_str = [_stringify_value(k, depth=depth + 1) for k in keys]
        values_str = [_stringify_value(v, depth=depth + 1) for v in values]
        keys_str = _to_str(keys_str, keys)
        values_str = _to_str(values_str, values)
        val_dict_str = {f"{k}:{v}" for k, v in zip(keys_str, values_str)}
        value_str = "\{"
        value_str += ",".join(val_dict_str)
        value_str += "\}"
    else:
        all_vals = [_stringify_value(v, depth=depth + 1) for v in value]
        value_str = ",".join(
            str(v)
            if not isinstance(v, str) or isinstance(ov, (set, tuple, list, dict))
            else f"'\"'\"'{v}'\"'\"'"
            for v, ov in zip(all_vals, value)
        )
        if isinstance(value, tuple):
            value_str = f"\({value_str},\)" if len(value) > 0 else "\(\)"
        else:
            value_str = f"[{value_str}]"

    return value_str


def _to_str(all_vals, value):
    return [
        str(v)
        if not isinstance(v, str) or isinstance(ov, (set, tuple, list, dict))
        else f"'\"'\"'{v}'\"'\"'"
        for v, ov in zip(all_vals, value)
    ]
